import { AudioContext, IAudioNode } from 'standardized-audio-context';
import { EventTarget } from '@web-media/event-target';
import { OpenPromise } from '@web-media/open-promise';

interface IDataChunk<ChunkMetadata> {
    duration: number;
    rawData: Uint8Array;
    metadata: ChunkMetadata;
    frames: number;
    readonly wrappedData: Uint8Array;
}
interface IAppendDataResult<ChunkMetadata> {
    consumed: number;
    /**
     * This value is only for debug purpose.
     */
    skipped?: number;
    data: IDataChunk<ChunkMetadata>;
}
declare abstract class MediaDeMuxAdapter<FileMetadata, ChunkMetadata> {
    abstract appendData(x: Uint8Array, isFirstChunk: boolean, isLastChunk: boolean): IAppendDataResult<ChunkMetadata> | null;
}

interface RawFrameHeader {
    type: 'raw';
    mpegVersion: number;
    mpegLayer: number;
    sampleRate: number;
    channelMode: number;
}
interface ParsedMetadata {
    type: 'parsed';
    mpegVersion: number;
    mpegLayer: number;
    sampleRate: number;
    channelMode: string;
}
interface SeekableParsedMetadata extends ParsedMetadata {
    start: number;
    end: number;
}
declare class Mp3DeMuxAdapter extends MediaDeMuxAdapter<{}, SeekableParsedMetadata[]> {
    appendData: (value: Uint8Array, isFirstChunk: boolean, isLastChunk: boolean) => {
        skipped: number;
        consumed: number;
        data: {
            rawData: Uint8Array;
            metadata: SeekableParsedMetadata[];
            duration: number;
            frames: number;
            readonly wrappedData: Uint8Array;
        };
    } | null;
    static validateHeader: (data: Uint8Array, i?: number) => boolean;
    static processFrameHeader: (data: Uint8Array, i?: number) => RawFrameHeader;
    static parseMetadata: (metadata: RawFrameHeader) => ParsedMetadata;
    static wrapChunk: (original: Uint8Array) => Uint8Array;
    /**
     * Frame length is length of a frame when compressed. It is calculated in
     * slots. One slot is 4 bytes long for Layer I, and one byte long for Layer II
     * and Layer III.
     */
    static getFrameLength: (data: Uint8Array, i: number, metadata: ParsedMetadata) => number;
}

declare class BinaryLoader {
    readonly url: string;
    readonly abortSignal: AbortSignal;
    constructor(url: string, abortSignal: AbortSignal);
    fetched: boolean;
    totalSize: number;
    downloadedSize: number;
    progress: number;
    buffer: Uint8Array | null;
    consume: (size: number) => Uint8Array | null;
    fetch(): AsyncGenerator<Uint8Array, void, unknown>;
}

declare class Clip<FileMetadata, ChunkMetadata> extends EventTarget {
    url: string;
    loop: boolean;
    readonly adapter: MediaDeMuxAdapter<FileMetadata, ChunkMetadata>;
    context: AudioContext;
    buffered: number;
    length: number;
    loaded: OpenPromise<boolean>;
    canPlayThough: OpenPromise<boolean>;
    controller: AbortController;
    loader: BinaryLoader;
    playing: boolean;
    ended: boolean;
    private _startTime;
    private _currentTime;
    private _chunks;
    private _contextTimeAtStart;
    private _pendingSourceStart;
    private _connected;
    private fadeTarget;
    private _gain;
    private _loadStarted;
    private _actualPlaying;
    get stuck(): boolean;
    private _currentSource;
    private _nextSource;
    private _currentGain;
    private _nextGain;
    private _audioBufferCache;
    constructor({ context, url, loop, volume, adapter, }: {
        context?: AudioContext;
        url: string;
        loop?: boolean;
        volume?: number;
        adapter: MediaDeMuxAdapter<FileMetadata, ChunkMetadata>;
    });
    /**
     * calculates if enough audio data is available to start playing and finish
     * downloading before running out.
     * If the conditions inside the function are met, it will resolve the
     * `canPlayThough` promise to `true`. Otherwise, it may not perform any
     * action.
     */
    private checkCanPlayThrough;
    private handleChunkReady;
    private handleChunkError;
    /**
     * Asynchronously loads and processes audio data, checking for readiness to
     * play through and handling errors accordingly.
     */
    buffer(): Promise<void>;
    connect(destination: IAudioNode<AudioContext>, output?: number, input?: number): this;
    disconnect(destination: IAudioNode<AudioContext>, output?: number, input?: number): void;
    _disconnectAllAndReplaceAudioContext(newAudioContext: AudioContext): void;
    dispose(): void;
    play(): Promise<unknown>;
    pause(): this;
    get currentTime(): number;
    set currentTime(currentTime: number);
    get duration(): number | null;
    get paused(): boolean;
    get volume(): number;
    set volume(volume: number);
    fade(startVolume: number, endVolume: number, duration: number): void;
    private stopFade;
    private trySetupAudioBufferCache;
    private audioBufferCacheHit;
    private advanceAudioBufferCache;
    private startPlay;
    private advanceAudioNodes;
    private resetAudioNodes;
    private onCurrentSourceEnd;
    private onBufferDecoded;
    private decodeChunk;
    private tryResumePlayback;
}

export { Clip, Mp3DeMuxAdapter };
//# sourceMappingURL=types.d.ts.map
